/************************************************************************
 *   Insert a new element into a singly linked, non-empty sorted list   *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>


// Data types

typedef double D;         // data values

struct L_Node {           // list nodes
  struct L_Node* next;
  D data;
};
typedef struct L_Node* L; // lists


// Code

void insert(L p, L e)
// p points to head of non-empty list
// e points to new element
{
  L s = p;
  L t = s->next;
  while (t != NULL) {
    if (e->data >= t->data) {
      s = t;
      t = s->next;
    }
    else break;
  }
  e->next = t;
  s->next = e;
}


/******************************************************************************

*** State signature
value sorts:      D
pointer sorts:    L
functions:        next:L->L, data:L->D
constants:        p,e,s,t:L
rigid constants:  NULL:L
rigid predicates: leq,lt:D,D->2
ghost constants:  p0,e0:L
ghost predicates: S0:L->2


*** Control flow graph
source: insert.dot
graph:  insert.png


*** Control locations l_i in CFG
l_0, l_1, l_2


*** Paths sigma_j through CFG
sigma_1: l_0->l_1 :: s = p; t = s->next;

sigma_2: l_1->l_1 :: (t != NULL) (e->data >= t->data) s = t; t = s->next;

sigma_3: l_1->l_2 :: (t != NULL) (e->data < t->data) e->next = t; s->next = e;

sigma_4: l_1->l_2 :: (t == NULL) e->next = t; s->next = e;


*** Path formulae pi_j for sigma_j
pi_1 :: s' = p &&
        t' = next(s') &&
        p' = p && e' = e && next' = next && data' = data

pi_2 :: t != NULL &&
        data(t) <= data(e) &&
        s' = t &&
        t' = next(s') &&
        p' = p && e' = e && next' = next && data' = data

pi_3 :: exists next_1:L->L .
          t != NULL &&
          data(e) < data(t) &&
          next_1(e) = t && (all x:L . x = e || next_1(x) = next(x)) &&
          next'(s) = e && (all x:L . x = s || next'(x) = next_1(x)) &&
          p' = p && e' = e && s' = s && t' = t && data' = data

pi_4 :: exists next_1:L->L .
          t = NULL &&
          next_1(e) = t && (all x:L . x = e || next_1(x) = next(x)) &&
          next'(s) = e && (all x:L . x = s || next'(x) = next_1(x)) &&
          p' = p && e' = e && s' = s && t' = t && data' = data


*** LPs defining shape types
P_List1[L; next:L->L, p,NULL:L; S:L->2] =
    {!S(NULL),
     S(p),
     all x:L . S(x) ==> next(x) != p,
     all x:L . S(x) && next(x) != NULL ==> S(next(x)),
     all x,y,z:L . S(x) && S(y) && S(z) && next(x) = z && next(y) = z ==>
                   x = y}

P_List1-[L; next:L->L, p,e,NULL:L; S:L->2] =
    P_List1[L; next,p,NULL; S] \cup {!S(e)}

P_List1+[L; next:L->L, p,e,NULL:L; S:L->2] =
    P_List1[L; next,p,NULL:L; S] \cup {S(e)}

P_SList1[D,L; next:L->L, data:L->D, p,NULL:L, <=:D,D->2; S:L->2] =
    P_List1[L; next,p,NULL; S] \cup
    {all x:L . S(x) && S(next(x)) ==> data(x) <= data(next(x))}

P_SList1-[D,L; next:L->L, data:L->D, p,e,NULL:L, <=:D,D->2; S:L->2] =
    P_SList1[L; next,data,p,NULL,<=; S] \cup P_List1-[L; next,p,e,NULL; S]

P_SList1+[D,L; next:L->L, data:L->D, p,e,NULL:L, <=:D,D->2; S:L->2] =
    P_SList1[L; next,data,p,NULL,<=; S] \cup P_List1+[L; next,p,e,NULL; S]


*** Shape invariants <phi_i,P_i> at l_i (with signature ext Delta_i)
Delta_0 :: S,E:L->2
phi_0   :: S = S0 && p = p0 && e = e0 &&
           e != NULL && data(p) <= data(e)
P_0     :: P_SList1-[D,L; next,data,p,e,NULL,<=; S]

Delta_1 :: S,E:L->2
phi_1   :: S = S0 && p = p0 && e = e0 &&
           e != NULL && data(p) <= data(e) &&
           S(s) && data(s) <= data(e) && next(s) = t && (t = NULL || S(t))
P_1     :: P_SList1-[D,L; next,data,p,e,NULL,<=; S]

Delta_2 :: S:L->2
phi_2   :: S = S0 \uplus {e0} && p = p0 && e = e0
P_2     :: P_SList1+[D,L; next,data,p,e,NULL,<=; S]


*** Shape effects epsilon_j for sigma_j
epsilon_1 :: S' = S
epsilon_2 :: S' = S
epsilon_3 :: S' = S \uplus {e}
epsilon_4 :: S' = S \uplus {e}

Notes:
1) The signature extensions Delta and Lambda are implicit:
   - Delta is the Delta of the shape invariant at the start of sigma_j,
   - Lambda is the Delta of the shape invariant at the end of sigma_j.

2) Back-and-forth totality for the shape effects epsilon_j (j = 1,2)
   follow from the following trivial conditions:
     (forth) all S:L->2 . exists S':L->2 . S' = S
     (back)  all S':L->2 . exists S:L->2 . S' = S

   Back-and-forth totality for the shape effects epsilon_j (j = 3,4)
   follows from the following less trivial conditions; notice that
   (forth) holds because P_1, the LP at the start location l_1, entails !S(e),
   and (back) holds because P_2', the LP at the end location l_2, entails
   S'(e') and pi_j entails e = e':
     (forth) all S:L->2 . P_1 ==> exists S':L->2 . S' = S \uplus {e}
     (back)  all S':L->2 . pi_j && P_2 ==> exists S:L->2 . S' = S \uplus {e}


*** Non-minimality formulae nMIN_i at l_i
nMIN_0 :: S^ \subseteq S &&
          !(S \subseteq S^)

nMIN_1 :: S^ \subseteq S &&
          !(S \subseteq S^)

nMIN_2 :: S^ \subsetq S &&
          !(S \subseteq S^)

Notes:
1) The signature extensions Delta are implicit:
   - Delta is the Delta of the shape invariant at l_i

2) The non-minimality formulae are completely determined by the signature.
   They are listed here only for completeness.

3) "S \subseteq S^" is a shorthand for "all x:L . S(x) ==> S^(x)",
   "!(S \subseteq S^)" is then short for "exists x:L . S(x) && !S^(x)".


*** Verification conditions:
VCs are the ones from the end of Section 3.2 of the paper, except that
when the conclusion of a VC is a big conjunction (eg. the clauses of
an LP) we split the VC into several, one for each conjectured conjunct.
This seems to suit some provers better than leaving the conjecture
in one big piece.

******************************************************************************/
